# Generated by Django 4.0.5 on 2022-07-31 22:50

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('urlid_graph', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='elementconfig',
            options={'verbose_name': 'Element configuration', 'verbose_name_plural': 'Element configurations'},
        ),
        migrations.AlterModelOptions(
            name='entity',
            options={'verbose_name': 'Entity', 'verbose_name_plural': 'Entities'},
        ),
        migrations.AlterModelOptions(
            name='objectrepository',
            options={'verbose_name': 'Object repository', 'verbose_name_plural': 'Object repositories'},
        ),
        migrations.AlterModelOptions(
            name='savedgraph',
            options={'verbose_name': 'Saved graph', 'verbose_name_plural': 'Saved graphs'},
        ),
        migrations.AlterModelManagers(
            name='entity',
            managers=[
            ],
        ),
        migrations.AlterUniqueTogether(
            name='objectrepository',
            unique_together=set(),
        ),
    ]
