# Generated by Django 4.0.6 on 2022-08-09 10:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('urlid_graph', '0005_alter_joblog_options_alter_logstep_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='elementconfig',
            name='config_type',
            field=models.PositiveSmallIntegerField(choices=[(1, 'entity'), (2, 'relationship'), (3, 'property'), (4, 'stats'), (5, 'custom')]),
        ),
    ]
